/*
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005-2014, Anthony Minessale II <anthm@freeswitch.org>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Anthony Minessale II <anthm@freeswitch.org>
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Anthony Minessale II <anthm@freeswitch.org>
 *
 * switch_jitterbuffer.h -- Audio/Video Jitter Buffer
 *
 */


#ifndef SWITCH_VIDDERBUFFER_H
#define SWITCH_VIDDERBUFFER_H

#include <switch_red.h>


typedef enum {
	SJB_QUEUE_ONLY = (1 << 0)
} switch_jb_flag_t;

typedef enum {
	SJB_VIDEO = 0,
	SJB_AUDIO,
	SJB_TEXT
} switch_jb_type_t;


SWITCH_BEGIN_EXTERN_C
SWITCH_DECLARE(switch_status_t) switch_jb_create(switch_jb_t **jbp, switch_jb_type_t type,
												 uint32_t min_frame_len, uint32_t max_frame_len, switch_memory_pool_t *pool,
												 ulp_policy_t *ulp_policy);
SWITCH_DECLARE(switch_status_t) switch_jb_set_frames(switch_jb_t *jb, uint32_t min_frame_len, uint32_t max_frame_len);
SWITCH_DECLARE(switch_status_t) switch_jb_peek_frame(switch_jb_t *jb, uint32_t ts, uint16_t seq, int peek, switch_frame_t *frame);
SWITCH_DECLARE(switch_status_t) switch_jb_get_frames(switch_jb_t *jb, uint32_t *min_frame_len, uint32_t *max_frame_len, uint32_t *cur_frame_len, uint32_t *highest_frame_len);
SWITCH_DECLARE(switch_status_t) switch_jb_destroy(switch_jb_t **jbp);
SWITCH_DECLARE(void) switch_jb_reset(switch_jb_t *jb);
SWITCH_DECLARE(uint32_t) switch_jb_get_fec_rcvd_total(switch_jb_t *jb);
SWITCH_DECLARE(uint32_t) switch_jb_get_fec_success(switch_jb_t *jb);
SWITCH_DECLARE(uint32_t) switch_jb_get_nack_success(switch_jb_t *jb);
SWITCH_DECLARE(void) switch_jb_debug_level(switch_jb_t *jb, uint8_t level);
SWITCH_DECLARE(void) switch_jb_set_nack_fec_combo(switch_jb_t *jb);
SWITCH_DECLARE(void) switch_jb_unset_nack_fec_combo(switch_jb_t *jb);
SWITCH_DECLARE(uint32_t) switch_jb_get_packets_per_frame(switch_jb_t *jb);
SWITCH_DECLARE(int) switch_jb_frame_count(switch_jb_t *jb);
SWITCH_DECLARE(int) switch_jb_poll(switch_jb_t *jb);
SWITCH_DECLARE(switch_status_t) switch_jb_put_packet(switch_jb_t *jb, switch_rtp_packet_t *packet, switch_size_t len, switch_bool_t fec);
SWITCH_DECLARE(switch_size_t) switch_jb_get_last_read_len(switch_jb_t *jb);
SWITCH_DECLARE(switch_status_t) switch_jb_get_packet(switch_jb_t *jb, switch_rtp_packet_t *packet, switch_size_t *len);
SWITCH_DECLARE(uint32_t) switch_jb_pop_nack(switch_jb_t *jb);
SWITCH_DECLARE(switch_status_t) switch_jb_get_packet_by_seq(switch_jb_t *jb, uint16_t seq, switch_rtp_packet_t *packet, switch_size_t *len);
SWITCH_DECLARE(void) switch_jb_set_session(switch_jb_t *jb, switch_core_session_t *session);
SWITCH_DECLARE(void) switch_jb_ts_mode(switch_jb_t *jb, uint32_t samples_per_frame, uint32_t samples_per_second);
SWITCH_DECLARE(void) switch_jb_set_flag(switch_jb_t *jb, switch_jb_flag_t flag);
SWITCH_DECLARE(void) switch_jb_set_rtt(switch_jb_t *jb, uint16_t rtt);
SWITCH_DECLARE(void) switch_jb_clear_flag(switch_jb_t *jb, switch_jb_flag_t flag);

SWITCH_DECLARE(void) switch_jb_set_ulp_policy(switch_jb_t *jb, ulp_policy_t *policy);
SWITCH_DECLARE(ulp_policy_t *) switch_jb_get_ulp_policy(switch_jb_t *jb);


/* Internal methods. */


/* ULP sender's callback. */

/**
 * Copy requested RTP into @rtp array passed to this callback.
 * Return the number of RTP packets associated with FEC of snb = @snb,
 * on @k leveel of protection. ULP owns the RTP packets and will free them.
 *
 * Makes call to jb_ulp_rtp_material(policy, snb, k, rtp, 1).
 *
 * This callback is called when ULP generates FEC header for { @policy, @snb, @k }
 * as requested by user in a call to
 *
 *		ulp_fec_pkt_t* ulp_gen_fec(ulp_policy_t *policy, uint8_t fec_n, uint16_t snb)
 *
 * If the number of RTP is smaller than needed to generate FEC packet, then ulp_gen_fec() will fail.
 */
SWITCH_DECLARE(uint8_t) jb_ulp_cb_rtp_material(ulp_policy_t *policy, uint16_t snb, uint8_t k, ulp_rtp_msg_t *rtp);


/* ULP receiver's callbacks. */

/**
 * Return copy of the RTP with sequence @seq. Set @payload_original_len to length of the packet's body. ULP owns it and will free it.
 */
SWITCH_DECLARE(ulp_rtp_pkt_t*) jb_ulp_cb_rtp_by_seq(ulp_policy_t *policy, uint16_t seq, uint16_t *payload_original_len);

/**
 * Return the number of FEC packets which protect @seq on level @k. Packets are put into @fecs table,
 * which must be of size ULP_FEC_MAX_N. ULP owns packets and will free them.
 */
SWITCH_DECLARE(uint8_t) jb_ulp_cb_find_fecs_for_seq_on_lvl(ulp_policy_t *policy, uint16_t seq, uint8_t k, ulp_rtp_msg_t *fecs); 

/**
 * Receiver's callback called on RTP packet recovery. Need all packets in a group less one.
 * Return copies of packets in the @rtp table. ULP owns them and calls free on them.
 * Makes call to jb_ulp_rtp_material(policy, ntohs(fec->hdr.snb), k, rtp, 1).
 */
SWITCH_DECLARE(uint8_t) jb_ulp_cb_rtp_cover(ulp_policy_t *policy, ulp_fec_pkt_t *fec, uint16_t fec_payload_len, uint8_t k, ulp_rtp_msg_t *rtp);

/**
 * Re-recovery procedure.
 *
 * Successful recovery of @seq makes other recoveries possible. This method recovers packets which are now possible to be recovered
 * thanks to successful recovery of @seq.
 *
 * @seq - sequence successfully recovered which triggered this re recovery (making other recoveried possible)
 * @seq_shift - packets to check for possible recovery start from @seq - @seq_shift
 * @seq_n - number of packets to check for possible recovery (@seq_shift packets starting from @seq - @seq_shift)
 * @rerecovered_packet - memory where to created new packet, passed as an external pointer due to recursive nature of this function
 * @recursive_call_limit - max number of recursive calls
 * @re-recovery_attempt - number of recovery attempts made as result of initial successful recovery of @seq
 * @ssrc - host ordered SSRC of the to-be-recovered-packet
 */
SWITCH_DECLARE(void) switch_jb_re_recover(ulp_policy_t *ulp_policy, uint16_t seq, uint16_t seq_shift, uint16_t seq_n, switch_rtp_packet_t *rerecovered_packet, uint32_t *recursive_call_limit, uint32_t *re_recovery_attempt_n, uint32_t ssrc);

/**
 * Check if any packets protected by this FEC are missing and recover them.
 * @ssrc - host ordered SSRC of the to-be-recovered-packet, if it gets successfully recovered this SSRC is assigned to it
 */
uint8_t jb_recover_any_packets(ulp_fec_pkt_t *fec, ulp_policy_t *fec_ulp_policy, uint32_t ssrc);


/* Insert Redundant Encoding into hash.
 *
 * Remove oldest entry if hash would exceed SWITCH_RED_MAX_BUF_SZ. Increase hash size and save the redundancy sequence as the oldest.
 *
 * @redundancy_len - total length of the redundancy
 * @red_seq - seq of the RED wrapper the FEC come with
 * @ssrc - host ordered SSRC of the to-be-recovered-packet
 */
SWITCH_DECLARE(switch_status_t) switch_jb_put_redundancy(switch_rtp_packet_t *redundancy, switch_size_t redundancy_len, ulp_policy_t *fec_ulp_policy, uint16_t red_seq, uint32_t ssrc);

SWITCH_DECLARE(void) switch_jb_handle_recovered_pkt(switch_jb_t *jb, switch_rtp_packet_t *recovered_packet, switch_size_t recovered_len);
SWITCH_DECLARE(int) switch_jb_check_seq_is_red_fec(switch_jb_t *jb, uint16_t seq);
SWITCH_DECLARE(int) switch_jb_check_seq_is_fec(switch_jb_t *jb, uint16_t seq);

SWITCH_END_EXTERN_C
#endif

/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4 noet:
 */
